/* geoclue-stumbler-window.c
 *
 * Copyright 2013 Red Hat, Inc.
 * Copyright 2024 Christopher Talbot
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

// A lot of code shamelessly taken from https://gitlab.freedesktop.org/geoclue/geoclue/-/blob/master/demo/where-am-i.c?ref_type=heads

#include "config.h"

#include "geoclue-stumbler-window.h"
#include "geoclue-stumbler-maps-page.h"
#include "geoclue-stumbler-path-page.h"

#include <config.h>

#include <stdlib.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <geoclue.h>
#include <shumate/shumate.h>

#define GEOCLUE_TIMEOUT_SECONDS 30
//From Geoclue gclue_web_source.c
#define SUBMISSION_ACCURACY_THRESHOLD 100

struct _GeoclueStumblerWindow
{
    AdwApplicationWindow  parent_instance;
    AdwToastOverlay *toast_overlay;

    GtkLabel     *connected_label;
    GtkLabel     *submit_label;
    GtkLabel     *latitude_label;
    GtkLabel     *longitude_label;
    GtkLabel     *accuracy_label;
    GtkLabel     *altitude_label;
    AdwActionRow *altitude_row;
    GtkLabel     *speed_label;
    AdwActionRow *speed_row;
    GtkLabel     *heading_label;
    AdwActionRow *heading_row;
    GtkLabel     *source_label;
    AdwActionRow *source_row;
    GtkLabel     *timestamp_label;
    AdwActionRow *timestamp_row;

    GtkWidget    *maps_page_bin;
    GtkWidget    *path_page_bin;

    GClueSimple *simple;
};

G_DEFINE_FINAL_TYPE (GeoclueStumblerWindow, geoclue_stumbler_window, ADW_TYPE_APPLICATION_WINDOW)

static gboolean
on_location_timeout (gpointer user_data)
{
    AdwDialog *dialog;
    GeoclueStumblerWindow *self = user_data;

    if (self->simple)
        return G_SOURCE_REMOVE;

    dialog = adw_alert_dialog_new (_("Alert"),
                                   _("Cannot connect to Geoclue. Please quit this app and enable Geoclue."));

    adw_alert_dialog_add_responses (ADW_ALERT_DIALOG (dialog),
                                    "close",  _("_Close"),
                                    NULL);

    adw_dialog_present (dialog, GTK_WIDGET (self));

    return G_SOURCE_REMOVE;
}

static void
print_location (GeoclueStumblerWindow *self)
{
    GClueLocation *location;
    gdouble altitude, speed, heading;
    g_autofree char *latitude = NULL;
    g_autofree char *longitude = NULL;
    g_autofree char *accuracy = NULL;

    GVariant *timestamp;
    const char *desc;

    location = gclue_simple_get_location (self->simple);
    g_debug ("New location:");
    g_debug ("Latitude:    %f°", gclue_location_get_latitude (location));
    g_debug ("Longitude:   %f°", gclue_location_get_longitude (location));
    g_debug ("Accuracy:    %f meters", gclue_location_get_accuracy (location));

    latitude = g_strdup_printf("%f°", gclue_location_get_latitude (location));
    gtk_label_set_label (self->latitude_label, latitude);

    longitude = g_strdup_printf("%f°", gclue_location_get_longitude (location));
    gtk_label_set_label (self->longitude_label, longitude);

    if (gclue_location_get_accuracy (location) < 0.99)
        accuracy = g_strdup(_("Exact"));
    else
        accuracy = g_strdup_printf(_("%.3f meters"), gclue_location_get_accuracy (location));

    gtk_label_set_label (self->accuracy_label, accuracy);

    altitude = gclue_location_get_altitude (location);
    if (altitude != -G_MAXDOUBLE) {
        g_autofree char *alt_label = NULL;
        g_debug ("Altitude:    %.2f meters", altitude);
        alt_label = g_strdup_printf("%.2f meters", altitude);
        gtk_label_set_label (self->altitude_label, alt_label);
        gtk_widget_set_visible (GTK_WIDGET(self->altitude_row), true);
    } else {
        gtk_widget_set_visible (GTK_WIDGET(self->altitude_row), false);
        g_debug ("Altitude:   -G_MAXDOUBLE");
    }

    speed = gclue_location_get_speed (location);
    g_debug ("Speed:       %f meters/second", speed);
    if (speed > 0) {
        g_autofree char *speed_label_text = NULL;
        speed_label_text = g_strdup_printf(_("%.4f meters/second"), speed);
        gtk_label_set_label (self->speed_label, speed_label_text);
        gtk_widget_set_visible (GTK_WIDGET(self->speed_row), true);
    } else
        gtk_widget_set_visible (GTK_WIDGET(self->speed_row), false);

    heading = gclue_location_get_heading (location);
    g_debug ("Heading:     %f°", heading);
    if (heading > 0.0001) {
        g_autofree char *heading_label_text = NULL;
        heading_label_text = g_strdup_printf("%f°", heading);
        gtk_label_set_label (self->heading_label, heading_label_text);
        gtk_widget_set_visible (GTK_WIDGET(self->heading_row), true);
    } else
        gtk_widget_set_visible (GTK_WIDGET(self->heading_row), false);

    desc = gclue_location_get_description (location);
    g_debug ("Description: %s", desc);
    if (desc != NULL && strlen (desc) > 0) {
        gtk_label_set_label (self->source_label, desc);
        gtk_widget_set_visible (GTK_WIDGET(self->source_row), true);
    } else
        gtk_widget_set_visible (GTK_WIDGET(self->source_row), false);

    timestamp = gclue_location_get_timestamp (location);
    if (timestamp) {
        g_autoptr(GDateTime) date_time = NULL;
        guint64 sec, usec;
        g_autofree gchar *str = NULL;

        g_variant_get (timestamp, "(tt)", &sec, &usec);

        /* Ignore usec, since it is not in the output format */
        date_time = g_date_time_new_from_unix_local ((gint64) sec);
        str = g_date_time_format (date_time, "%c");

        g_debug ("Timestamp:   %s", str);
        gtk_label_set_label (self->timestamp_label, str);
        gtk_widget_set_visible(GTK_WIDGET(self->timestamp_row), true);
    } else
        gtk_widget_set_visible(GTK_WIDGET(self->timestamp_row), false);

    if (desc != NULL && strstr(desc, "GPS") != NULL && gclue_location_get_accuracy (location) <= SUBMISSION_ACCURACY_THRESHOLD) {
        gtk_label_set_label (self->submit_label, _("Yes"));

        geoclue_stumbler_path_page_set_gps_lock (GEOCLUE_STUMBLER_PATH_PAGE (self->path_page_bin), TRUE);
        geoclue_stumbler_path_page_update_page (GEOCLUE_STUMBLER_PATH_PAGE (self->path_page_bin),
                                                gclue_location_get_latitude (location),
                                                gclue_location_get_longitude (location),
                                                gclue_location_get_speed (location),
                                                gclue_location_get_altitude (location));

        geoclue_stumbler_maps_page_update_marker (GEOCLUE_STUMBLER_MAPS_PAGE (self->maps_page_bin),
                                                  gclue_location_get_latitude (location),
                                                  gclue_location_get_longitude (location),
                                                  gclue_location_get_heading (location));

    } else {
        gtk_label_set_label (self->submit_label, _("No"));
        geoclue_stumbler_path_page_set_gps_lock (GEOCLUE_STUMBLER_PATH_PAGE (self->path_page_bin), FALSE);
    }

}

static void
on_simple_ready (GObject      *source_object,
                 GAsyncResult *res,
                 gpointer      user_data)
{
    GeoclueStumblerWindow *self = user_data;
    GError *error = NULL;

    self->simple = gclue_simple_new_with_thresholds_finish (res, &error);
    if (error != NULL) {
        g_critical ("Failed to connect to GeoClue2 service: %s", error->message);

        on_location_timeout (self);
    }

    gtk_label_set_label (self->connected_label, "Connected");
    print_location (self);

    g_signal_connect_swapped (self->simple,
                              "notify::location",
                              G_CALLBACK (print_location),
                              self);
}

static void
geoclue_stumbler_window_finalize (GObject *object)
{
  GeoclueStumblerWindow *self = (GeoclueStumblerWindow *)object;
  g_clear_object (&self->simple);

  G_OBJECT_CLASS (geoclue_stumbler_window_parent_class)->finalize (object);
}


static void
geoclue_stumbler_window_class_init (GeoclueStumblerWindowClass *klass)
{
    GObjectClass   *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    gtk_widget_class_set_template_from_resource (widget_class, "/org/kop316/stumbler/geoclue-stumbler-window.ui");
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, connected_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, submit_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, latitude_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, longitude_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, accuracy_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, altitude_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, altitude_row);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, speed_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, speed_row);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, heading_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, heading_row);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, source_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, source_row);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, timestamp_label);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, timestamp_row);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, maps_page_bin);
    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, path_page_bin);

    gtk_widget_class_bind_template_child (widget_class, GeoclueStumblerWindow, toast_overlay);

    object_class->finalize = geoclue_stumbler_window_finalize;
}

static void
geoclue_stumbler_window_init (GeoclueStumblerWindow *self)
{
    GClueAccuracyLevel accuracy_level = GCLUE_ACCURACY_LEVEL_EXACT;
    gint time_threshold = 0;
    /*
    Shumate doesn't provide an init() function
    https://gitlab.gnome.org/GNOME/libshumate/-/issues/64
    */
    g_type_ensure (SHUMATE_TYPE_SIMPLE_MAP);
    g_type_ensure (GEOCLUE_STUMBLER_TYPE_MAPS_PAGE);
    g_type_ensure (GEOCLUE_STUMBLER_TYPE_PATH_PAGE);

    gtk_widget_init_template (GTK_WIDGET (self));

    gclue_simple_new_with_thresholds ("geoclue-where-am-i",
                                      accuracy_level,
                                      time_threshold,
                                      0,
                                      NULL,
                                      on_simple_ready,
                                      self);

    g_timeout_add_seconds (GEOCLUE_TIMEOUT_SECONDS, on_location_timeout, self);

    self->simple = NULL;
    geoclue_stumbler_path_page_set_toast_overlay (GEOCLUE_STUMBLER_PATH_PAGE (self->path_page_bin),
                                                  self->toast_overlay);

    geoclue_stumbler_path_page_set_maps_page (GEOCLUE_STUMBLER_PATH_PAGE (self->path_page_bin),
                                              self->maps_page_bin);

    geoclue_stumbler_maps_page_set_path_page (GEOCLUE_STUMBLER_MAPS_PAGE (self->maps_page_bin),
                                              self->path_page_bin);

}
